/*
 * msize: display sizes of MINIX/ST executable.
 * 
 * Copyright (c) 2010, 2011 Hans Ottevanger
 *	All rights reserved.
 *
 * This version is endian and wordsize independent.
 *
 */

#include <stdint.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

struct mnx_hdr {
	uint8_t  a_magic[2];
	uint8_t  a_flags;
	uint8_t  a_cpu;
	uint8_t  a_hdrlen;
	uint8_t  a_unused;
	uint16_t a_version;
	uint32_t a_text;
	uint32_t a_data;
	uint32_t a_bss;
	uint32_t a_entry;
	uint32_t a_total;
	uint32_t a_syms;
};

#define A_MAGIC0      0x01
#define A_MAGIC1      0x03
#define BADMAG(X)     ((X).a_magic[0] != A_MAGIC0 ||(X).a_magic[1] != A_MAGIC1)

#define A_UZP   0x01    /* unmapped zero page (pages) */
#define A_EXEC  0x10    /* executable */
#define A_SEP   0x20    /* separate I/D */

static uint32_t getuint32(FILE *);
static uint16_t getuint16(FILE *);
static uint8_t getuint8(FILE *);

static void fatal(const char *, ...);
static void size(const char *);

static char *progname;

int main(int argc, char *argv[])
{
	int i;

	progname = argv[0];

	if (argc == 1) {
		size("a.out");
		exit(EXIT_SUCCESS);
	}

	for (i = 1; i < argc; i++)
		size(argv[i]);

	exit(EXIT_SUCCESS);
}

static void size(const char *filename)
{
	FILE *file;
	struct mnx_hdr mexec;
	int separate;
	uint32_t dynam, allmem;

	static int heading = 0;

	if ((file = fopen(filename, "r")) == NULL) {
		fprintf(stderr, "%s: can't open %s\n", progname, filename);
		return;
	}

	mexec.a_magic[0] = getuint8(file);
	mexec.a_magic[1] = getuint8(file);
	mexec.a_flags = getuint8(file);
	mexec.a_cpu = getuint8(file);
	mexec.a_hdrlen = getuint8(file);
	mexec.a_unused = getuint8(file);
	mexec.a_version = getuint16(file);
	mexec.a_text = getuint32(file);
	mexec.a_data = getuint32(file);
	mexec.a_bss = getuint32(file);
	mexec.a_entry = getuint32(file);
	mexec.a_total = getuint32(file);
	mexec.a_syms = getuint32(file);

	if (BADMAG(mexec)) {
		fprintf(stderr, "%s: %s: file format not recognized\n", progname, filename);
		fclose(file);
		return;
	}

	separate = (mexec.a_flags & A_SEP ? 1 : 0);
	dynam = mexec.a_total - mexec.a_text - mexec.a_data - mexec.a_bss;
	if (separate)
		dynam += mexec.a_text;
	allmem = (separate ? mexec.a_total + mexec.a_text : mexec.a_total);
	if (heading++ == 0)
		printf("   text\t   data\t    bss\t  stack\t memory\n");
	printf("%7u\t%7u\t%7u\t%7u\t%7u\t%s\n",
        	mexec.a_text, mexec.a_data, mexec.a_bss, dynam, allmem, filename);

	fclose(file);

	return;
}

static uint32_t
getuint32(FILE *input)
{
	uint32_t result;

	result = getuint8(input);
	result = result << 8 | getuint8(input);
	result = result << 8 | getuint8(input);
	result = result << 8 | getuint8(input);

	return result;
}

static uint16_t
getuint16(FILE *input)
{
	uint16_t result;

	result = getuint8(input);
	result = result << 8 | getuint8(input);

	return result;
}

static uint8_t
getuint8(FILE *input)
{
	int result;

	if ((result = getc(input)) == EOF)
		fatal("read error\n");

	return result;
}

static void
fatal(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "%s: ", progname);
	vfprintf(stderr, fmt, ap);
	va_end(ap);

	exit(EXIT_FAILURE);
}
